/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    CarClassifier.java
 *    Copyright (C) 2015 Zbigniew Grudzinski
 *    
 *    original file: JCBA.java
 *	  Copyright (C) 2004 Stefan Mutter
 */
package weka.classifiers.rules.car;

import static weka.confirmationMeasures.ConfirmationMeasuresConstants.ABSOLUTE;
import static weka.confirmationMeasures.ConfirmationMeasuresConstants.NAN_VALUE;
import static weka.confirmationMeasures.ConfirmationMeasuresConstants.PERCENTAGE;
import static weka.confirmationMeasures.ConfirmationMeasuresConstants.ZERO_VALUE;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Vector;

import weka.associations.AbstractAssociator;
import weka.associations.Associator;
import weka.associations.CARuleMiner;
import weka.associations.ItemSet;
import weka.classifiers.AbstractClassifier;
import weka.classifiers.rules.car.utils.Stopwatch;
import weka.core.AdditionalMeasureProducer;
import weka.core.Attribute;
import weka.core.Capabilities;
import weka.core.Capabilities.Capability;
import weka.core.CapabilitiesHandler;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.RevisionUtils;
import weka.core.SelectedTag;
import weka.core.Tag;
import weka.core.TechnicalInformation;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import weka.filters.Filter;
import weka.filters.supervised.attribute.Discretize;

/**
 * @author Zbigniew Grudzinski
 */

public class CarClassifier extends AbstractClassifier implements OptionHandler,
		AdditionalMeasureProducer, CapabilitiesHandler, TechnicalInformationHandler {

	/** For serialization */
	static final long serialVersionUID = -4039170500265315498L;

	/** The mining algorithm. */
	private CARuleMiner m_assoc = new weka.associations.CarApriori();

	/** The instances. */
	private Instances m_instances;

	/** Watch for mining runtime behavior. */
	private Stopwatch m_miningWatch;

	/** The mining time. */
	private double m_miningTime;

	/** The filter to discretize attribute values. */
	private Discretize m_filter;

	/**
	 * Flag indicating if class association rules should be part of the
	 * classifier output.
	 */
	private boolean m_printAssoc = false;
	
	/**
	 * Name of the directory where serialized CAR rule sets are stored 
	 */
	private String m_carRuleSetDir = ".";

	/** List of all rules. */
	private ArrayList<CarRule> m_rules;

	/** Conflict resolution strategy comparator */
	private MeasuresComparator m_conflictResolutionStrategyComparator;
	
	/** Rules number reduction strategy comparator */
	private MeasuresComparator m_rulesNumberReductionStrategyComparator;

	/** Rules limit according to chosen rules limit data type */
	private int m_rulesLimit = 1000;

	/** The division by zero strategy for confirmation measure */
	private int m_divisionByZeroStrategy = NAN_VALUE;

	/** Rules limit data type */
	private int m_rulesLimitDataType = ABSOLUTE;
	
	/** Should rule set statistics be saved to file */
	private boolean m_saveRuleSetStatistics = false;
	
	/**
	 * Name of the directory where rule set statistics files will be stored.
	 */
	private String m_ruleSetStatisticsDir = ".";

	/** The division by zero strategy for confirmation measure */
	private static final Tag[] DIVISION_BY_ZERO_STRATEGY = {
			new Tag(NAN_VALUE, "confirmation measure set to NaN"),
			new Tag(ZERO_VALUE, "confirmation measure set to 0 if a * d - b * c == 0 otherwise NaN") };

	/** Rules limit data type */
	private static final Tag[] RULES_LIMIT_DATA_TYPE = { new Tag(ABSOLUTE, "absolute value"),
			new Tag(PERCENTAGE, "percentage value") };

	/**
	 * Returns default capabilities of the base associator.
	 * 
	 * @return the capabilities of the base classifier
	 */
	@Override
	public Capabilities getCapabilities() {
		Capabilities result;

		if (getCarMiner() != null && getCarMiner() instanceof CapabilitiesHandler) {
			result = ((CapabilitiesHandler) getCarMiner()).getCapabilities();
			result.disable(Capability.NO_CLASS);
		} else {
			result = new Capabilities(this);
			result.disableAll();
		}

		// set dependencies
		for (Capability cap : Capability.values()) {
			result.enableDependency(cap);
		}

		result.setOwner(this);

		return result;
	}

	/**
	 * Returns an instance of a TechnicalInformation object, containing detailed
	 * information about the technical background of this class, e.g., paper
	 * reference or book this class is based on.
	 * 
	 * @return the technical information about this class
	 */
	@Override
	public TechnicalInformation getTechnicalInformation() {
		TechnicalInformation result;
		TechnicalInformation additional;

		result = new TechnicalInformation(Type.INPROCEEDINGS);
		result.setValue(Field.AUTHOR, "Bing Liu and Wynne Hsu and Yiming Ma");
		result.setValue(Field.TITLE, "Integrating Classification and Association Rule Mining");
		result.setValue(Field.BOOKTITLE,
				"Fourth International Conference on Knowledge Discovery and Data Mining");
		result.setValue(Field.YEAR, "1998");
		result.setValue(Field.PAGES, "80-86");
		result.setValue(Field.PUBLISHER, "AAAI Press");

		additional = result.add(Type.INPROCEEDINGS);
		additional.setValue(Field.AUTHOR, "W. Li and J. Han and J.Pei");
		additional
				.setValue(Field.TITLE,
						"CMAR: Accurate and Efficient Classification Based on Multiple Class-Association Rules");
		additional.setValue(Field.BOOKTITLE, "ICDM'01");
		additional.setValue(Field.YEAR, "2001");
		additional.setValue(Field.PAGES, "369-376");
		// additional.setValue(Field.PUBLISHER, "AAAI Press");

		return result;
	}

	/**
	 * Gets a description of the CMCAR algorithm
	 * 
	 * @return a description of the CMCAR algorithm
	 */
	public String globalInfo() {

		return  "Implementation of CMCAR algorithm (Confirmation Measures Class Association Rules) which is based on CBA algoritm with added confirmation measures."
				+ "Original CBA authors and publication " + getTechnicalInformation().toString();
	}

	/**
	 * Returns an enumeration describing the available options
	 * 
	 * @return an enumeration of all the available options
	 */
	@Override
	public Enumeration<Option> listOptions() {

		Vector<Option> newVector = new Vector<Option>(10);

		newVector.addElement(new Option(
				"\tFull class name of class association rule miner to include, followed\n"
						+ "\tby scheme options.\n" + "\tMust produce class association rules.\n"
						+ "\teg: \"weka.associations.Apriori\"", "A", 1,
				"-A <class association rule miner specification>"));

		newVector.addElement(new Option(
				"\t If set class association rules are also part of the output(default no).\n",
				"V", 0, "-V"));
		
		newVector.addElement(new Option(
				"\t Name of the directory where serialized CAR rule sets are stored (default is current working directory \".\").\n",
				"-CAR_RULE_SET_DIR", 1, "-CAR_RULE_SET_DIR <directory_name>"));

		newVector
				.addElement(new Option(
						"\tConflict resolution strategy. List of comma separated measures, which are used to resolve conflicts when more than one rule with different consequences could match specific instance."
						+ " Listed measures are applied in sequence when previous measure haven't solved conflict."
						+ " Posible values are confirmation measures: D | M | S | N | C | F | A | Z | C1 | C2 | C3 | C4 but also Support (sup), Confidence (conf), rule Length (length)\n"
								+ "\t(default is conf)", "CONFLICT_RESOLUTION_ST", 1,
						"-CONFLICT_RESOLUTION_ST <d | m | s | n | c | f | a | z | c1 | c2 | c3 | c4 | sup | conf | length>"));
		
		newVector
				.addElement(new Option(
						"\tRules number reduction strategy. List of comma separated measures, which are used to reduce number of rules."
						+ " Listed measures are applied in sequence when sorting rule set."
						+ " Posible values are confirmation measures: D | M | S | N | C | F | A | Z | C1 | C2 | C3 | C4 but also Support (sup), Confidence (conf), rule Length (length)\n"
								+ "\t(default is conf)", "RULES_NUMBER_REDUCTION_ST", 1,
						"-RULES_NUMBER_REDUCTION_ST <d | m | s | n | c | f | a | z | c1 | c2 | c3 | c4 | sup | conf | length>"));

		newVector
				.addElement(new Option(
						"\tRules limit. Maximum number of rules if algorithm generates more rules they will be truncated to this number.\n"
								+ "\t(default is 1000)", "RULES_LIMIT", 1, "-RULES_LIMIT <num>"));
		newVector
				.addElement(new Option(
						"\tRules limit data type. Clarifies wheter rules limit parameter indicates absolute value or percentage value of all rules generated by algoritm. Possible values are absolute | percentage\n"
								+ "\t(default is abs)", "RULES_LIMIT_DATA_TYPE", 1,
						"-RULES_LIMIT_DATA_TYPE <abs | %>"));

		newVector
			.addElement(new Option(
				"\tDivision by zero in confirmation measure strategy. Possible values are: confirmation measure coerced to NaN | confirmation measure coerced to 0 if a * d - b * c == 0 otherwise NaN.\n"
						+ "\t(default is nan)", "DIV_BY_ZERO_ST", 1,
				"-DIV_BY_ZERO_ST <nan | zero>"));
		
		newVector.addElement(new Option(
				"\t If set rule set statistics will be saved to specified directory (default no).\n",
				"SAVE_RULE_SET_STATS", 0, "-SAVE_RULE_SET_STATS"));

		newVector.addElement(new Option(
				"\t Name of the directory where where rule set statistics files will be stored (default is current working directory \".\").\n",
				"-RULE_SET_STATS_DIR", 1, "-RULE_SET_STATS_DIR <directory_name>"));

		return newVector.elements();
	}

	/**
	 * Parses a given list of options. Valid options are:
	 * <p>
	 * 
	 * 
	 * -A Class Association Rule Miner String <br>
	 * Class Association Rule Miner String should contain the full class name of
	 * a scheme included for selection followed by options to the Class
	 * Association Rule Miner.
	 * <p>
	 * 
	 * @param options
	 *            the list of options as an array of strings
	 * @exception Exception
	 *                if an option is not supported
	 */
	@Override
	public void setOptions(String[] options) throws Exception {

		String assocString = Utils.getOption('A', options);
		if (assocString.length() > 0) {
			// throw new
			// Exception("class association rule miner must be specified");

			String[] assocSpec = Utils.splitOptions(assocString);

			if (assocSpec.length == 0) {
				throw new Exception("Invalid class association rule miner specification string");
			}
			String assocName = assocSpec[0];
			assocSpec[0] = "";
			Associator toUse = (AbstractAssociator.forName(assocName, assocSpec));
			if (!(toUse instanceof CARuleMiner)) {
				throw new Exception(
						"Association Rule Miner has to be able to mine classification association rules");
			}
			setCarMiner((CARuleMiner) toUse);
		}

		m_printAssoc = Utils.getFlag('V', options);
		
		String carRuleSetDir= Utils.getOption("CAR_RULE_SET_DIR", options);
		if (carRuleSetDir != null && !carRuleSetDir.isEmpty()) {
			m_carRuleSetDir = carRuleSetDir;
		}

		String value = Utils.getOption("CONFLICT_RESOLUTION_ST", options);
		if (value != null) {
			setConflictResolutionStrategy(value.toUpperCase());
		} else {
			throw new IllegalArgumentException("Invalid conflict resolution strategy");
		}
		
		value = Utils.getOption("RULES_NUMBER_REDUCTION_ST", options);
		if (value != null) {
			setRulesNumberReductionStrategy(value.toUpperCase());
		} else {
			throw new IllegalArgumentException("Invalid rules number reduction strategy");
		}

		value = Utils.getOption("RULES_LIMIT_DATA_TYPE", options);
		if (value.toLowerCase().equals("abs")) {
			setRulesLimitDataType(new SelectedTag(ABSOLUTE, RULES_LIMIT_DATA_TYPE));
		} else if (value.toLowerCase().equals("%")) {
			setRulesLimitDataType(new SelectedTag(PERCENTAGE, RULES_LIMIT_DATA_TYPE));
		} else {
			throw new IllegalArgumentException("Invalid rules limit data type");
		}

		value = Utils.getOption("RULES_LIMIT", options);
		try {
			int numValue = Integer.parseInt(value);
			if (numValue < 0) {
				throw new IllegalArgumentException("Invalid rules limit");
			}
			setRulesLimit(numValue);
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException("Invalid rules limit");
		}

		value = Utils.getOption("DIV_BY_ZERO_ST", options);
		if (value.toLowerCase().equals("nan")) {
			setDivisionByZeroStrategy(new SelectedTag(NAN_VALUE, DIVISION_BY_ZERO_STRATEGY));
		} else if (value.toLowerCase().equals("zero")) {
			setDivisionByZeroStrategy(new SelectedTag(ZERO_VALUE, DIVISION_BY_ZERO_STRATEGY));
		} else {
			throw new IllegalArgumentException(
					"Invalid zero division strategy for confirmation measure");
		}
		
		m_saveRuleSetStatistics = Utils.getFlag("SAVE_RULE_SET_STATS", options);
		
		String ruleSetStatisticsDir = Utils.getOption("RULE_SET_STATS_DIR", options);
		if (ruleSetStatisticsDir != null && !ruleSetStatisticsDir.isEmpty()) {
			if (m_saveRuleSetStatistics) {
				m_ruleSetStatisticsDir = ruleSetStatisticsDir;
			} else {
				System.err.println("SAVE_RULE_SET_STATS option not set. RULE_SET_STATS_DIR option ignored.");
			}
		}

		Utils.checkForRemainingOptions(options);
	}

	/**
	 * Gets the current settings of the Classifier.
	 * 
	 * @return an array of strings suitable for passing to setOptions
	 */
	@Override
	public String[] getOptions() {

		String[] options = new String[19];
		int current = 0;

		options[current++] = "-A";
		options[current++] = "" + getAssocSpec();

		if (m_printAssoc) {
			options[current++] = "-V";
		}
		
		options[current++] = "-CAR_RULE_SET_DIR";
		options[current++] = "" + m_carRuleSetDir;

		options[current++] = "-CONFLICT_RESOLUTION_ST";
		options[current++] = "" + getConflictResolutionStrategy();
		
		options[current++] = "-RULES_NUMBER_REDUCTION_ST";
		options[current++] = "" + getRulesNumberReductionStrategy();

		options[current++] = "-RULES_LIMIT";
		options[current++] = "" + m_rulesLimit;

		options[current++] = "-RULES_LIMIT_DATA_TYPE";
		options[current++] = "" + m_rulesLimitDataType;

		options[current++] = "-DIV_BY_ZERO_ST";
		options[current++] = "" + m_divisionByZeroStrategy;
		
		if (m_saveRuleSetStatistics) {
			options[current++] = "-SAVE_RULE_SET_STATS";
		}
		
		options[current++] = "-RULE_SET_STATS_DIR";
		options[current++] = "" + m_ruleSetStatisticsDir;

		while (current < options.length) {
			options[current++] = "";
		}
		return options;
	}

	/**
	 * Gets the options for the class association rule miner
	 * 
	 * @return string with the options
	 */
	protected String getAssocSpec() {

		if (m_assoc instanceof OptionHandler) {
			return m_assoc.getClass().getName() + " "
					+ Utils.joinOptions(((OptionHandler) m_assoc).getOptions());
		}
		return m_assoc.getClass().getName();
	}

	/**
	 * Gets the tipText for this option.
	 * 
	 * @return the tipText for this option.
	 */
	public String carMinerTipText() {

		return "The class association rule miner with its options.";
	}

	/**
	 * Sets the class association rule miner
	 * 
	 * @param assoc
	 *            the class association rule miner
	 */
	public void setCarMiner(CARuleMiner assoc) {

		m_assoc = assoc;
	}

	/**
	 * Gets the class association rule miner
	 * 
	 * @return the class association rule miner
	 */
	public CARuleMiner getCarMiner() {

		return m_assoc;
	}

	/**
	 * Gets the tipText for the specified option.
	 * 
	 * @return the tipText for the specified option.
	 */
	public String printAssoctTipText() {

		return "Enables/Disables the output of the association algorithm.";
	}

	/**
	 * Sets whether or not class association rules should be part of the
	 * classifier output
	 * 
	 * @param flag
	 *            boolean flag
	 */
	public void setPrintAssoc(boolean flag) {

		m_printAssoc = flag;
	}

	/**
	 * Gets whether or not class association rules should be part of the
	 * classifier output
	 * 
	 * @return true if class association rules should be part of the classifier
	 *         output, false otherwise
	 */
	public boolean getPrintAssoc() {

		return m_printAssoc;
	}

	/**
	 * Gets the tipText for the specified option.
	 * 
	 * @return the tipText for the specified option.
	 */
	public String carRuleSetDirTipText() {

		return "Name of the directory where serialized CAR rule sets are stored.";
	}

	/**
	 * Sets name of the directory where serialized CAR rule sets are stored
	 * 
	 * @param directoryName
	 *            name of the directory
	 */
	public void setCarRuleSetDir(String directoryName) {

		m_carRuleSetDir = directoryName;
	}

	/**
	 * Gets name of the directory where serialized CAR rule sets are stored
	 * 
	 * @return Name of the directory where serialized CAR rule sets are stored.
	 */
	public String getCarRuleSetDir() {

		return m_carRuleSetDir;
	}
	
	/**
	 * Sets the conflict resolution strategy comparator
	 * 
	 * @param resolutionMeasures
	 *            comma separated String of ComparisonMeasures used to resolve conflicts
	 */
	public void setConflictResolutionStrategy(String resolutionMeasures) {
		ArrayList<ComparisonMeasure> comparisionMeasures = new ArrayList<ComparisonMeasure>();
		for(String measure : resolutionMeasures.split(",")) {
			comparisionMeasures.add(ComparisonMeasure.valueOf(measure));
		}
		m_conflictResolutionStrategyComparator = new MeasuresComparator(comparisionMeasures.toArray(new ComparisonMeasure[comparisionMeasures.size()]));
	}

	/**
	 * Gets the conflict resolution strategy
	 * 
	 * @return the comma separated String of ComparisonMeasures used to resolve conflicts
	 */
	public String getConflictResolutionStrategy() {
		return m_conflictResolutionStrategyComparator.getComparisonMeasuresCommaSeparatedList();
	}
	
	
	/**
	 * Sets the rules number reduction strategy comparator
	 * 
	 * @param reductionMeasures
	 *            comma separated String of ComparisonMeasures used to reduce number or rules
	 */
	public void setRulesNumberReductionStrategy(String reductionMeasures) {
		ArrayList<ComparisonMeasure> comparisionMeasures = new ArrayList<ComparisonMeasure>();
		for(String measure : reductionMeasures.split(",")) {
			comparisionMeasures.add(ComparisonMeasure.valueOf(measure));
		}
		m_rulesNumberReductionStrategyComparator = new MeasuresComparator(comparisionMeasures.toArray(new ComparisonMeasure[comparisionMeasures.size()]));
	}
	
	/**
	 * Gets the rules number reduction strategy
	 * 
	 * @return the comma separated String of ComparisonMeasures used to reduce number or rules
	 */
	public String getRulesNumberReductionStrategy() {
		return m_rulesNumberReductionStrategyComparator.getComparisonMeasuresCommaSeparatedList();
	}

	/**
	 * Sets the rules limit according to chosen rule limit data type
	 * 
	 * @param rulesLimit
	 *            the rules limit
	 */
	public void setRulesLimit(int rulesLimit) {
		m_rulesLimit = rulesLimit;
	}

	/**
	 * Gets the maximum number of rules if algorithm generates more rules they
	 * will be truncated to this number
	 * 
	 * @return the rules limit according to chosen rules limit data type
	 */
	public int getRulesLimit() {
		return m_rulesLimit;
	}

	/**
	 * Sets the division by zero strategy
	 * 
	 * @param divisionByZeroStrategy
	 *            the division by zero strategy
	 */
	public void setDivisionByZeroStrategy(SelectedTag divisionByZeroStrategy) {
		if (divisionByZeroStrategy.getTags() == DIVISION_BY_ZERO_STRATEGY) {
			m_divisionByZeroStrategy = divisionByZeroStrategy.getSelectedTag().getID();
		}
	}

	/**
	 * Gets the division by zero strategy for confirmation measure.
	 * 
	 * @return the division by zero strategy
	 */
	public SelectedTag getDivisionByZeroStrategy() {
		return new SelectedTag(m_divisionByZeroStrategy, DIVISION_BY_ZERO_STRATEGY);
	}

	/**
	 * Sets the rules limit data type
	 * 
	 * @param rulesLimitDataType
	 *            rules limit data type
	 */
	public void setRulesLimitDataType(SelectedTag rulesLimitDataType) {
		if (rulesLimitDataType.getTags() == RULES_LIMIT_DATA_TYPE) {
			m_rulesLimitDataType = rulesLimitDataType.getSelectedTag().getID();
		}
	}

	/**
	 * Gets the rules limit data type
	 * 
	 * @return rules limit data type
	 */
	public SelectedTag getRulesLimitDataType() {
		return new SelectedTag(m_rulesLimitDataType, RULES_LIMIT_DATA_TYPE);
	}

	
	/**
	 * Gets the tipText for the specified option.
	 * 
	 * @return the tipText for the specified option.
	 */
	public String saveRuleSetStatisticsTipText() {

		return "Should rule set statistics be saved to file.";
	}

	/**
	 * Sets boolean value, whether rule set statistics should be saved to file.
	 * 
	 * @param saveRuleSetStatistics 
	 *          boolean value, whether rule set statistics should be saved to file
	 */
	public void setSaveRuleSetStatistics(boolean saveRuleSetStatistics) {

		m_saveRuleSetStatistics = saveRuleSetStatistics;
	}

	/**
	 * Gets information whether rule set statistics should be saved to file.
	 * 
	 * @return Information whether rule set statistics should be saved to file.
	 */
	public boolean getSaveRuleSetStatistics() {

		return m_saveRuleSetStatistics;
	}
	
	/**
	 * Gets the tipText for the specified option.
	 * 
	 * @return the tipText for the specified option.
	 */
	public String ruleSetStatisticsDirTipText() {

		return "Name of the directory where rule set statistics files will be stored.";
	}

	/**
	 * Sets name of the directory where rule set statistics files will be stored.
	 * 
	 * @param directoryName
	 *            name of the directory
	 */
	public void setRuleSetStatisticsDir(String directoryName) {

		m_ruleSetStatisticsDir = directoryName;
	}

	/**
	 * Gets name of the directory where rule set statistics files will be stored.
	 * 
	 * @return Name of the directory where rule set statistics files will be stored.
	 */
	public String getRuleSetStatisticsDir() {

		return m_ruleSetStatisticsDir;
	}
	
	/**
	 * Generates the classifier.
	 * 
	 * @param newInstances
	 *            set of instances serving as training data
	 * @exception Exception
	 *                if the classifier has not been generated successfully
	 */
	@Override
	public void buildClassifier(Instances newInstances) throws Exception {

		ArrayList<Object>[] allRules;
		String relationName = newInstances.relationName().split("[^a-zA-Z0-9]+")[0];
		Path carRuleSetDirPath =  Paths.get(m_carRuleSetDir);
		String carRuleSetDir;
		if (!Files.exists(carRuleSetDirPath)) {
			throw new Exception(
					"Directory where serialized CAR rule sets are stored doesn't exist");
		}
		
		try {
			carRuleSetDir = carRuleSetDirPath.toRealPath().toString();
		} catch (SecurityException e) {
			throw new Exception(
					"Not enough permissions to read and write from/to directory where serialized CAR rule sets are stored");
		}
		
		String serializedFileName = relationName + "_" + getAssocSpec().replaceAll("weka\\.associations\\.|\\s+","") + ".rules";
		File file = new File(carRuleSetDir, serializedFileName);
		
		boolean needsDisc = false;
		for (int i = 0; i < newInstances.numAttributes(); i++) {
			if (i != newInstances.classIndex() && newInstances.attribute(i).isNumeric()) {
				needsDisc = true;
				break;
			}
		}

		Instances instances = newInstances;
		if (needsDisc) {
			m_filter = new Discretize();
			m_filter.setInputFormat(newInstances);
			instances = Filter.useFilter(newInstances, m_filter);
		} else {
			m_filter = null;
		}
		m_instances = instances;
		getCapabilities().testWithFail(instances);
		
		ArrayList<CarRule> rules;
		
		if (!file.exists()) {
		
			m_miningWatch = new Stopwatch();
			m_miningWatch.start();
			m_assoc.setClassIndex(m_instances.classIndex() + 1);
			allRules = m_assoc.mineCARs(m_instances);
			m_miningTime = m_miningWatch.getSeconds();
			m_miningWatch.stop();
	
			rules = new ArrayList<CarRule>(allRules[0].size());
			CarRule carRule;
			for (int i = 0; i < allRules[0].size(); i++) {
				carRule = new CarRule(((ItemSet) allRules[0].get(i)).items(),
						((ItemSet) allRules[1].get(i)).itemAt(0));
				carRule.computeMeasures(instances, m_divisionByZeroStrategy);
				rules.add(carRule);
			}			
			
			try (FileOutputStream outputFileStream = new FileOutputStream(file)) {
				try (ObjectOutputStream objectOutputStream = new ObjectOutputStream(outputFileStream)) {
					objectOutputStream.writeObject(rules);
				}
			}
		} else {
			
			// Association Algorithm hasn't been executed
			m_printAssoc = false;
			
			try (FileInputStream inputFileStream = new FileInputStream(file)) {
				try (ObjectInputStream objectInputStream = new ObjectInputStream(inputFileStream)) {
					@SuppressWarnings("unchecked")
					ArrayList<CarRule> serializedRules = (ArrayList<CarRule>) objectInputStream.readObject();
					rules = serializedRules;
				}
			}
		}
		
		if (m_saveRuleSetStatistics) {
			String dirName = m_ruleSetStatisticsDir + File.separator + "carMinerRules";
			saveRuleSetStatistics(rules, dirName, relationName + ".all");
		}
		
		int maxRules = rules.size();
		if (m_rulesLimitDataType == ABSOLUTE) {
			maxRules = m_rulesLimit > 0 ? Math.min(m_rulesLimit, maxRules) : maxRules;
		} else if (m_rulesLimitDataType == PERCENTAGE) {
			maxRules = Math.min((int) ((m_rulesLimit / 100.0) * maxRules), maxRules);
		}
		
		ArrayList<CarRule> limitRuleSet = new ArrayList<CarRule>(rules);
		
		rules.sort(m_conflictResolutionStrategyComparator);
		
		// optimization if rules list size is equal to maxRules we can skip whole reduction strategy part
		if (rules.size() == maxRules) {
			m_rules = rules;
		} else {
			
			limitRuleSet.sort(m_rulesNumberReductionStrategyComparator);
			
			ArrayList<CarRule> finalRuleSet = new ArrayList<CarRule>(maxRules);
			HashSet<Integer> chosenRulesIds = new HashSet<Integer>(maxRules);
			for (int i = 0; i < maxRules; i++) {
				chosenRulesIds.add(limitRuleSet.get(i).getRuleId());
			}
			
			for(CarRule rule : rules) {
				if (chosenRulesIds.contains(rule.getRuleId())) {
					finalRuleSet.add(rule);
					
					// optimization if finalRuleSet has maxRules elements we don't have check next rule in rules list
					if (finalRuleSet.size() == maxRules) {
						break;
					}
				}
			}
			
			m_rules = finalRuleSet;
		}
		
		if (m_saveRuleSetStatistics) {
			String dirName = m_ruleSetStatisticsDir + File.separator + 
					getConflictResolutionStrategy() + File.separator +  getRulesNumberReductionStrategy();
			saveRuleSetStatistics(m_rules, dirName, relationName);
		}
		
		m_instances = new Instances(instances, 0);

	}

	/**
	 * Classifies an instance
	 * 
	 * @param instance
	 *            the instance
	 * @throws Exception
	 *             exception if instance cannot be classified
	 * @return the predicted class
	 */
	@Override
	public double classifyInstance(Instance instance) throws Exception {

		double result = Double.NaN;
		double[] distribution = innerDistributionForInstance(instance);

		if (!Utils.eq(Utils.sum(distribution), 0)) {
			result = Utils.maxIndex(distribution);
		}
		return result;
	}

	/**
	 * Calculates inner distribution for given instance
	 * @param instance
	 * 			single instance
	 * @return distribution for given instance
	 */
	public double[] innerDistributionForInstance(Instance instance) {
		double[] distribution = new double[m_instances.numClasses()];

		for (int i = 0; i < m_rules.size(); ++i) {
			if (m_rules.get(i).covers(instance)) {
				distribution[(int) m_rules.get(i).getConsequent()] = 1;
				break;
			}
		}

		if (!Utils.eq(Utils.sum(distribution), 0)) {
			Utils.normalize(distribution);
		}

		return distribution;
	}
	
	/**
	 * Saves rule set statistics to file fileName in directory dirName.
	 * @param rules 
	 * 			list of rules.
	 * @param dirName 
 * 				name of the directory where filename should be created.
	 * @param fileName name of the file where rule set statistics will be saved. 
	 */
	private void saveRuleSetStatistics(ArrayList<CarRule> rules, String dirName, String fileName) {
		if (rules.isEmpty()) {
			return;
		}
		
		try {			
			Path dirPath = Paths.get(dirName);
			if (!Files.exists(dirPath)) {
			    Files.createDirectories(dirPath);
			}
			
			Path filePath = Paths.get(dirName, fileName);
			
			if (!Files.exists(filePath)) {
				
				try (BufferedWriter writer = 
							new BufferedWriter(new FileWriter(filePath.toFile()))) {
					for(CarRule rule : rules) {
						writer.write(rule.getRuleStatistics());
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Prints the rules
	 * 
	 * @return the rules
	 */
	@Override
	public String toString() {

		int classIndex = m_instances.classIndex();
		Attribute classAttribute = m_instances.attribute(classIndex);
		Instances instancesWithoutClass = new Instances(m_instances);
		instancesWithoutClass.setClassIndex(-1);
		instancesWithoutClass.deleteAttributeAt(classIndex);

		StringBuffer text = new StringBuffer();

		if (m_printAssoc) {
			text.append(m_assoc.toString());
			text.append("\n\n");
		}

		text.append("Mining Time in sec.: " + measureMiningTime() + "\n");
		text.append("\n");

		text.append("Number of Rules: " + m_rules.size() + "\n");
		
		for (int i = 0; i < m_rules.size(); i++) {

			String n = Utils.doubleToString((double) i + 1,
					(int) (Math.log(m_rules.size()) / Math.log(10) + 1), 0);

			text.append(n).append(". ")
					.append(m_rules.get(i).toString(instancesWithoutClass, classAttribute));

			text.append('\n');
		}
		return text.toString();
	}

	/**
	 * Lists all additional statistics that are available
	 * 
	 * @return list of the names of the additional measures
	 */
	@Override
	public Enumeration<String> enumerateMeasures() {
		Vector<String> newVector = new Vector<String>(5);
		newVector.addElement("measureMiningTime");
		newVector.addElement("measurePruningTime");
		newVector.addElement("measureNumMinedRules");
		newVector.addElement("measureNumRulesAfterOptPruning");
		newVector.addElement("measureNumClassRules");
		return newVector.elements();
	}

	/**
	 * Gets the additional statistics
	 * 
	 * @param additionalMeasureName
	 *            the name of the additional measure
	 * @return the value of the measure
	 */
	@Override
	public double getMeasure(String additionalMeasureName) {
		if (additionalMeasureName.compareTo("measureMiningTime") == 0) {
			return measureMiningTime();
		} else {
			throw new IllegalArgumentException(additionalMeasureName + " not supported (CBA)");
		}
	}

	/**
	 * Gets the mining time
	 * 
	 * @return the mining time in seconds
	 */
	public double measureMiningTime() {

		return m_miningTime;
	}

	/**
	 * Returns the revision string.
	 * 
	 * @return the revision
	 */
	@Override
	public String getRevision() {
		return RevisionUtils.extract("$Revision: 10209 $");
	}

	/**
	 * Main method for testing this class.
	 * 
	 * @param argv
	 *            the options
	 */
	public static void main(String[] argv) {
		runClassifier(new CarClassifier(), argv);
	}
}
