package weka.classifiers.rules.car;

import java.io.Serializable;
import java.util.Arrays;
import java.util.EnumMap;

import weka.classifiers.rules.Rule;
import weka.confirmationMeasures.ConfirmationMeasure;
import weka.confirmationMeasures.ConfirmationMeasureFactory;
import weka.core.Attribute;
import weka.core.Instance;
import weka.core.Instances;

/**
 * Class implementing CAR rule (class association rule)
 * @author Zbigniew Grudzinski
 */
public class CarRule extends Rule implements Serializable {
	private static final long serialVersionUID = 7839030997018897507L;
	private static int counter;
	
	private int m_ruleId;
	private int[] m_conditions;
	private int m_consequent;
	
	private double m_coverage;
	private double m_uncoverage;
	private double m_support;
	private double m_confidence;
	
	/* contingency table for rule E -> H
	 *  H | ~H
	E | a | c
	~E| b | d
	*/
	
	/**
	 * a - Number of objects, that satisfies both the premise and the conclusion of the rule
	 */
	private double m_satisfiesConditionsAndConsequent;
	
	/**
	 * b - Number of objects not satisfying rule the premise, but satisfying its conclusion
	 */
	private double m_satisfiesOnlyConsquent;
	
	/**
	 * c - Number of objects satisfying the premise, but not the conclusion of the rule
	 */
	private double m_satisfiesOnlyConditions;
	
	/**
	 * d - Number of objects, that do not satisfy neither the premise not the conclusion of the rule
	 */
	private double m_satisfiesNeitherConditionsNorConsquent;
	
	private EnumMap<ComparisonMeasure, Double> m_measures;
	
	/**
	 * Default constructor - creates empty rule
	 */
	public CarRule() {
		m_ruleId = counter;
		counter++;
		
		m_measures = new EnumMap<ComparisonMeasure, Double>(ComparisonMeasure.class);
	}
	
	/**
	 * Constructor which creates a rule based upon premise and consequence given as its parameters
	 * @param premise 
	 * 			premise of the rule
	 * @param consequence 
	 * 			consequence of the rule
	 */
	public CarRule(int[] premise, int consequence) {
		this();
		m_conditions = premise;
		m_consequent = consequence;
	}

	/**
	 * Checks if rule covers given instance
	 * @param datum 
	 * 			single instances
	 */
	@Override
	public boolean covers(Instance datum) {
		for (int atribute_index = 0; atribute_index < m_conditions.length; atribute_index++) {
			if (!datum.isMissing(atribute_index)) {
				double value = datum.value(atribute_index);
				double ruleCondition = m_conditions[atribute_index];
				if (ruleCondition != -1 && value != ruleCondition) {
					return false;
				}
			}
		}
		return true;
	}
	
	/**
	 * Gets unique id of the rule
	 * @return Unique id of the rule
	 */
	public int getRuleId() {
		return m_ruleId;
	}
	
	/**
	 * Gets rule conditions (its premise)
	 * @return array of rule conditions (its premise)
	 */
	public int[] getConditions() {
		return m_conditions;
	}
	
	/**
	 * Gets EnumMap which contains value of all confirmation measures, rule support, confidence and legth (of premise)
	 * @return EnumMap which contains value of all confirmation measures, rule support, confidence and legth (of premise)
	 */
	public EnumMap<ComparisonMeasure, Double> getMeasures() {
		return m_measures;
	}

	/**
	 * Gets rule consequent (class)
	 * @return rule consequent (its class)
	 */
	@Override
	public double getConsequent() {
		return m_consequent;
	}

	/**
	 * getRevision operation is unsupported
	 */
	public String getRevision() {
		throw new UnsupportedOperationException("getRevision is forbidden");
	}

	/**
	 * grow operation is unsupported
	 */
	@Override
	public void grow(Instances data) throws Exception {
		throw new UnsupportedOperationException("grow is forbidden");
	}

	/**
	 * Has the rule got any antecedents
	 */
	@Override
	public boolean hasAntds() {
		return size() != 0;
	}

	/**
	 * Size of the rule
	 */
	@Override
	public double size() {
		int antdsCount = 0;
		for (int i = 0; i < m_conditions.length; i++) {
			if (m_conditions[i] != -1) {
				antdsCount++;
			}
		}
		return antdsCount;
	}

	/**
	 * Computers value of all Bayesian confirmation measures and rule support, confidence, length (of premise) and stores them in m_measures EnumMap
	 * @param instances 
	 * 			list of instances
	 * @param divisionByZeroStrategyId 
 * 				id of division by zero strategy
	 */
	public void computeMeasures(Instances instances, int divisionByZeroStrategyId)
	{
		computeSimpleStats(instances);
		
		// create all confirmation measures objects, compute their measures and put them into measures HashMap
		String[] confirmationMeasuresNames = {"D", "M", "S", "N", "C", "F", "Z", "A", "C1", "C2", "C3", "C4"};
		for (String measureName : confirmationMeasuresNames) {
			ConfirmationMeasureFactory factory = new ConfirmationMeasureFactory(measureName, divisionByZeroStrategyId);
			ConfirmationMeasure confirmationMeasure = factory.createMeasure();
			confirmationMeasure.computeMeasure(m_satisfiesConditionsAndConsequent, m_satisfiesOnlyConsquent, m_satisfiesOnlyConditions, m_satisfiesNeitherConditionsNorConsquent);
			
			m_measures.put(ComparisonMeasure.valueOf(measureName), confirmationMeasure.getMeasureValue());
		}
		
		m_measures.put(ComparisonMeasure.SUP, m_support);
		m_measures.put(ComparisonMeasure.CONF, m_confidence);
		m_measures.put(ComparisonMeasure.LENGTH, size());
	}
	
	/**
	 * Creates string containing values of confirmation measures and simple representation of decision rule without information about attribute names
	 * @return confirmation measures values and simple representation of decision rule
	 */
	@Override
	public String toString() {
		StringBuffer text = new StringBuffer();
		for (ComparisonMeasure measure : m_measures.keySet()) {
			text.append(measure.toString() + "=");
			text.append(m_measures.get(measure));
			text.append(";");
		}
		text.append(Arrays.toString(m_conditions));
		text.append(" ==> " + m_consequent);
		
		return text.toString();
	}
	
	/**
	 * Creates string representation of decision rule
	 * @param instancesWithoutClass 
	 * 				list of instances without class attribute
	 * @param classAttribute 
	 * 				attribute marked as class
	 * @return string representation of decision rule
	 */
	public String toString(Instances instancesWithoutClass, Attribute classAttribute) {
	    StringBuffer text = new StringBuffer();

	    for (int i = 0; i < instancesWithoutClass.numAttributes(); i++) {
	      if (m_conditions[i] != -1) {
	        text.append(instancesWithoutClass.attribute(i).name() + '=');
	        text.append(instancesWithoutClass.attribute(i).value(m_conditions[i]) + ' ');
	      }
	    }
	    
	    text.append("==> ");
	   
	    text.append(classAttribute.name() + '=');
        text.append(classAttribute.value(m_consequent) + ' ');
	    
	    return text.toString();
	}
	
	/**
	 * Create rule statistics
	 * @return string value containing rule statistics
	 */
	public String getRuleStatistics()
	{
		double confirmation = m_satisfiesConditionsAndConsequent * m_satisfiesNeitherConditionsNorConsquent - m_satisfiesOnlyConsquent * m_satisfiesOnlyConditions;
		
		StringBuffer text = new StringBuffer();
		text.append("ruleId=\t" + m_ruleId);
		text.append("\ta=\t" + m_satisfiesConditionsAndConsequent);
		text.append("\tb=\t" + m_satisfiesOnlyConsquent);
		text.append("\tc=\t" + m_satisfiesOnlyConditions);
		text.append("\td=\t" + m_satisfiesNeitherConditionsNorConsquent);
		
		for (ComparisonMeasure measure : m_measures.keySet()) {
			text.append("\t" + measure + "=\t" + m_measures.get(measure));
		}
		
		text.append("\t" + (confirmation > 0 ? "confirmation" : (confirmation < 0 ? "disconfirmation" : "none")) + "\t");
		text.append('\n');
		
		return text.toString();
	}
	
	/**
	 * Computes parameters required to calculate value of Bayesian confirmation measures
	 * @param instances 
	 * 			list of instances
	 */
	private void computeSimpleStats(Instances instances) {
		for (int i = 0; i < instances.numInstances(); i++) {
			Instance datum = instances.instance(i);
			double weight = datum.weight();
			if (covers(datum)) {
				m_coverage += weight; // Coverage
				if ((int) datum.classValue() == (int) getConsequent()) {
					m_satisfiesConditionsAndConsequent += weight; // True positives
				} else {
					m_satisfiesOnlyConditions += weight; // False positives
				}
			} else {
				m_uncoverage += weight;
				if ((int) datum.classValue() != (int) getConsequent()) {
					m_satisfiesNeitherConditionsNorConsquent += weight; // True negatives
				} else {
					m_satisfiesOnlyConsquent += weight; // False negatives
				}
			}
		}
		
		m_support = m_satisfiesConditionsAndConsequent / instances.numInstances();
		
		m_confidence = m_satisfiesConditionsAndConsequent / m_coverage;
	}
}
