package weka.classifiers.rules.car;

import java.io.Serializable;
import java.util.Comparator;

/**
 * Implementation of Comparator interface used to compare rules in rule set
 * @author Zbigniew Grudzinski
 */
public class MeasuresComparator implements Comparator<CarRule>, Serializable {

	private static final long serialVersionUID = -4333408430040699642L;

	private final double epsilon = 0.000001;
	
	private ComparisonMeasure[] m_sortOrder;
	
	/**
	 * Constructor used to initialize comparator sort order
	 * @param order 
	 * 			list of comparison measures
	 */
	public MeasuresComparator(ComparisonMeasure... order) {
		m_sortOrder = order;
	}

	/**
	 * Creates comma separated string representation of comparison measures used in comparator
	 * @return comma separated string representation of comparison measures used in comparator
	 */
	public String getComparisonMeasuresCommaSeparatedList() {
		String comparisonList = "";
		for (ComparisonMeasure order : m_sortOrder) {
			comparisonList += order.toString() + ",";
		}

		// remove last comma
		if (comparisonList.length() > 0) {
			comparisonList = comparisonList.substring(0, comparisonList.length() - 1);
		}
		
		return comparisonList;
	}
	
	/**
	 * Compares two rules using comparison measures in descending order
	 * @param rule1
	 * 			first rule to compare
	 * @param rule2 
	 * 			second rule to compare
	 * @return value less than 0 if measure1 is greater than measure2 or measure2 is NaN;
	 * 		   value greater than 0 if measure1 is less than measure2 or measure1 is NaN;
	 *         0 if rules are equal on all comparison measures values
	 */
	@Override
	public int compare(CarRule rule1, CarRule rule2) {
		for (ComparisonMeasure order : m_sortOrder) {
			Double measure1 = rule1.getMeasures().get(order);
			Double measure2 = rule2.getMeasures().get(order);
			
			if (measure1 == null || measure2 == null) {
				throw new NullPointerException("One or both of the measures are null");
			}
			
			if (Double.isNaN(measure1)) {
				return 1;
			} else if (Double.isNaN(measure2)) {
				return -1;
			} else if (Math.abs(measure1 - measure2) >= epsilon) {
				// descending order
				return Double.compare(measure2, measure1);
			}
		}
		
		// default behavior if both measures are equal on all ComaprisonMeasures values
		return 0;
	}

}
